import importlib.util
import sys
from inspect import getmembers, isbuiltin, isfunction
from pathlib import Path
from typing import Callable


def find_local_function(module) -> Callable | None:
    # getmembers returns a list of tuples
    for name, item in getmembers(module):
        # Check if it's a function but not a builtin
        if isfunction(item) and not isbuiltin(item):
            # Check if function is defined in this module by comparing its code location
            # Also check if the function's module is the same or a submodule
            if hasattr(item, "__code__") and (
                item.__module__ == module.__name__
                or item.__module__.startswith(f"{module.__name__}.")
            ):
                return item

    # Nothing found
    return None


def load_module(module_name: str, path: str | Path):
    path = Path(path)

    # Handle directory case
    if path.is_dir():
        init_file = path / "__init__.py"
        if not init_file.is_file():
            raise FileNotFoundError(f"'{init_file}' does not exist")
        target_path = init_file
    # Handle file case
    elif path.is_file():
        target_path = path
    else:
        raise ValueError(f"Path '{path}' is neither a file nor directory")

    # Create a module spec from the location
    spec = importlib.util.spec_from_file_location(module_name, str(target_path))
    if spec is None:
        raise ImportError(f"Cannot find spec for module '{module_name}' at '{path}'")

    # Create and execute module
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def handle(path: str) -> Callable:
    module = load_module("stardust.app", path)
    return find_local_function(module)



import contextlib
from inspect import signature, iscoroutinefunction
from typing import Any, Callable

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.routing import Route


class Stardust:
    def __init__(self, fun: Callable = None, port=8000, debug=False):
        self.debug: bool = debug
        self.app = None
        self.port = port
        self.method = fun

        if not self.method:
            print("File must contain at least one local function.")
            exit(1)

    async def __wrap_response(self, request: Request) -> Any:
        call_if_params = lambda method, req: (
            method(req) if len(signature(method).parameters) > 0 else method()
        )

        response = (
            await call_if_params(self.method, request)
            if iscoroutinefunction(self.method)
            else call_if_params(self.method, request)
        )

        if not issubclass(response.__class__, Response):
            # by default, we reply with json...
            response = JSONResponse(response)
        return response

    async def lifespan(self, app):
        async with contextlib.AsyncExitStack() as stack:
            print(f"Stardust listening on {self.port} 🎉")
            yield
            print(f"Shutting down Stardust on {self.port} 💥")
    
    def build(self):
        middlewares = [
            Middleware(CORSMiddleware, allow_origins=["*"]),
        ]

        routes = [
            Route("/", self.__wrap_response, methods=["GET", "POST"]),
        ]

        self.app = Starlette(
            debug=self.debug,
            middleware=middlewares,
            routes=routes,
            lifespan=self.lifespan,
        )

        return self.app



import random
from asyncio import sleep


async def load_model():
    """
    Loading the ML Model
    :return:
    """
    await sleep(2)


async def predict():
    """
    Emulate the load of the model and a prediction call
    :return: Number from 0.0 to 1.0
    """
    await load_model()
    return random.random()


async def serve():
    prediction = await predict()
    return {"Hello": "World", "prediction": prediction}



import inspect
import os
import sys
import tempfile

import pytest

from stardust.file_handler import handle


@pytest.fixture
def clean_module():
    if "stardust.app" in sys.modules:
        del sys.modules["stardust.app"]
    yield


def __write_app(tmp_path, app: str):
    p = tmp_path / tempfile.NamedTemporaryFile(suffix=".py").name
    p.write_text(app)
    return p.as_posix()


def test_simple_file(tmp_path, clean_module):
    app = """
    async def serve(req):
        return {"hello": "world"}
    """.strip()
    path = __write_app(tmp_path, app)

    fun = handle(path)
    assert inspect.isfunction(fun)
    assert fun.__name__ == "serve"


def test_import_lib(tmp_path, clean_module):
    app = """
from asyncio import sleep

async def serve(req):
    await sleep(1)
    return {"hello": "world1"}
    """.strip()
    path = __write_app(tmp_path, app)

    fun = handle(path)
    assert inspect.isfunction(fun)
    assert fun.__name__ == "serve"


def test_multiple_app(tmp_path, clean_module):
    app = """
async def serve1(req):
    return {"hello": "world1"}

async def serve2(req):
    return {"hello": "world2"}
    """.strip()
    path = __write_app(tmp_path, app)

    fun = handle(path)
    assert inspect.isfunction(fun)
    assert fun.__name__ == "serve1"


def test_module(tmp_path, clean_module):
    app_path = os.path.join(os.getcwd(), "tests", "test_file_handler")
    fun = handle(app_path)
    assert fun is not None
    assert inspect.isfunction(fun)
    assert fun.__name__ == "serve_1"



import pytest
from starlette.requests import Request
from starlette.responses import Response, PlainTextResponse
from starlette.testclient import TestClient

from stardust import Stardust


def test_json_app():
    async def serve():
        return {"hello": "world"}

    app = Stardust(fun=serve).build()
    client = TestClient(app)
    response = client.get("/")
    assert response.status_code == 200
    assert "hello" in response.json()


def test_json_app_with_request():
    async def serve(req):
        return {"hello": "world"}

    app = Stardust(fun=serve).build()
    client = TestClient(app)
    response = client.get("/")
    assert response.status_code == 200
    assert "hello" in response.json()


def test_custom_status_code():
    async def serve(req):
        return Response(status_code=204)

    app = Stardust(fun=serve).build()
    client = TestClient(app)
    response = client.get("/")
    assert response.status_code == 204


def test_response_plain_text():
    async def serve(req):
        return PlainTextResponse("Hello World!")

    app = Stardust(fun=serve).build()
    client = TestClient(app)
    response = client.get("/")
    assert response.status_code == 200
    assert "Hello" in response.text


def test_query_params():
    async def serve(req: Request):
        hello = req.query_params.get("hello", "all")
        return {"hello": hello}

    app = Stardust(fun=serve).build()
    client = TestClient(app)
    response = client.get("/?hello=world")
    assert response.status_code == 200
    assert "world" in response.json()["hello"]


def test_post_payload():
    async def echo(req: Request):
        body = await req.json()
        return body

    json = {"hello": "world"}

    app = Stardust(fun=echo).build()
    client = TestClient(app)
    response = client.post("/", json=json)
    assert response.status_code == 200
    assert json == response.json()


def test_headers_response():
    async def serve():
        return Response(status_code=200, headers={"x-custom-header": "hello world"})

    app = Stardust(fun=serve).build()
    client = TestClient(app)
    response = client.get("/")
    assert response.status_code == 200
    assert "x-custom-header" in response.headers
    assert "hello" in response.headers["x-custom-header"]


def test_empty_file(capsys):
    captured = capsys.readouterr()
    with pytest.raises(SystemExit) as e:
        app = Stardust().build()
        assert captured.out == "File must contain at least one local function."



import argparse

import uvicorn

from .file_handler import handle
from .stardust import Stardust


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file",
        nargs="?",
        default="app.py",
        help="Python file or directory with at least one async function that returns a string, object or Response.",
    )

    parser.add_argument(
        "--file",
        dest="file",
        help="Python file or directory with at least one async function that returns a string, object or Response.",
    )

    parser.add_argument(
        "--port",
        metavar="P",
        type=int,
        default=8000,
        help="Port number where the framework will listen for incoming requests.",
    )

    parser.add_argument(
        "--host",
        default="0.0.0.0",
        help="Host address to bind the server to.",
    )
    parser.add_argument(
        "--log-level",
        choices=["critical", "error", "warning", "info", "debug"],
        default="error",
        help="Logging level for the server.",
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Enable debug option.",
    )

    args = parser.parse_args()
    fun = handle(args.file)
    app = Stardust(fun=fun, port=args.port, debug=args.debug).build()
    uvicorn.run(
        app,
        host=args.host,
        port=args.port,
        access_log=args.debug,
        log_level=args.log_level,
    )
